import enum
import gzip
import logging
import os
import random
import re
import subprocess
from collections import defaultdict
from operator import itemgetter

import gtfparse
import pysam
import tabix
from sortedcontainers import SortedList

logger = logging.getLogger("utils")


class NGSFileType(enum.Enum):
    FASTQ = 1
    SAM = 2
    BAM = 3


class NGSFile:
    def __init__(self, filename, store_qualities=False):
        self.filename = filename
        self.store_qualities = store_qualities
        self.basename = os.path.basename(self.filename)
        self.ext = ".".join(self.basename.split(".")[1:])
        self.readmode = "r"
        self.gzipped = False

        if (
            self.ext.endswith(".gz")
            or self.ext.endswith(".bgz")
            or self.ext.endswith(".bam")
        ):
            self.readmode = "rb"
            self.gzipped = True

        if (
            self.ext.endswith("fastq")
            or self.ext.endswith("fq")
            or self.ext.endswith("fastq.gz")
            or self.ext.endswith("fq.gz")
        ):
            self.filetype = NGSFileType.FASTQ
            if self.gzipped:
                self.handle = gzip.open(self.filename, mode=self.readmode)
            else:
                self.handle = open(self.filename, mode=self.readmode)
        elif self.ext.endswith("sam"):
            self.filetype = NGSFileType.SAM
            self.handle = pysam.AlignmentFile(self.filename, self.readmode)
        elif self.ext.endswith("bam"):
            self.filetype = NGSFileType.BAM
            self.handle = pysam.AlignmentFile(self.filename, self.readmode)
        else:
            raise RuntimeError(
                "Could not determine NGS file type: {}".format(self.filename)
            )

        logger.debug("Opened NGS file '{}' as {}".format(self.filename, self.filetype))
        logger.debug("Gzipped: {}".format(self.gzipped))
        logger.debug("Readmode: {}".format(self.readmode))

    def __iter__(self):
        self.read_num = 0
        return self

    def __next__(self):
        query_name = None
        query = None
        read_group = None
        quality = None
        if self.filetype == NGSFileType.FASTQ:
            query_name = self.handle.readline().strip()
            if not query_name or query_name == "":
                raise StopIteration()

            query = self.handle.readline().strip()
            _plusline = self.handle.readline().strip()
            quality_string = self.handle.readline().strip()

            if self.gzipped:
                query_name = query_name.decode("utf-8")
                query = query.decode("utf-8")
                if self.store_qualities:
                    quality_string = quality_string.decode("utf-8")

            quality = []
            if self.store_qualities:
                for char in quality_string:
                    # PHRED+33 decoding
                    ascii_code = ord(char)
                    quality.append(ascii_code - 33)

            if query_name.startswith("@"):
                query_name = query_name[1:]

        elif self.filetype == NGSFileType.SAM or self.filetype == NGSFileType.BAM:
            read = next(self.handle)
            query_name = read.query_name
            query = read.query_alignment_sequence
            read_group = read.get_tag("RG")
            if self.store_qualities:
                quality = read.query_alignment_qualities

        self.read_num += 1

        if self.store_qualities:
            if read_group:
                return {
                    "query_name": query_name,
                    "query": query,
                    "read_group": read_group,
                    "quality": quality,
                }
            else:
                return {"query_name": query_name, "query": query, "quality": quality}
        else:
            if read_group:
                return {
                    "query_name": query_name,
                    "query": query,
                    "read_group": read_group,
                }
            else:
                return {"query_name": query_name, "query": query}


def sort_gff(filename):
    sorted_gff_name_tmp = filename
    ext = sorted_gff_name_tmp.split(".")[-1]
    gzipped = False
    if ext.endswith("gz"):
        gzipped = True
        sorted_gff_name_tmp = ".".join(sorted_gff_name_tmp.split(".")[:-1])
        ext = sorted_gff_name_tmp.split(".")[-1]
        handle = gzip.open(filename, "r")
    else:
        handle = open(filename, "r")
    sorted_gff_name_tmp = ".".join(
        sorted_gff_name_tmp.split(".")[:-1]
    )  # remove gtf/gff/gff3 ext
    sorted_gff_name = ".".join([sorted_gff_name_tmp, "sorted", ext])
    compressed_gff_name = ".".join([sorted_gff_name, "gz"])

    # Test if this was done in a previous run
    already_created = True
    try:
        # pytabix outputs unwanted text when file doesn't exist
        if os.path.isfile(compressed_gff_name):
            tabixed = tabix.open(compressed_gff_name)
            tabixed.queryi(0, 100, 200)
        else:
            raise FileNotFoundError
    except (FileNotFoundError, tabix.TabixError):
        already_created = False

    if already_created:
        logger.warning(f"Found and reusing {compressed_gff_name}.")
        return compressed_gff_name

    entries = []
    header_lines = []
    header_parsed = False  # only keep comments at start of file
    for line in handle:
        if gzipped:
            line = line.decode("utf-8")
        line = line.strip()

        if not line:
            continue
        if not header_parsed and line[0] == "#":
            header_lines.append(line)
            continue
        if line[0] == "#":
            continue
        header_parsed = True

        [
            seqname,
            source,
            feature,
            start,
            end,
            score,
            strand,
            frame,
            attributes,
        ] = line.split("\t")

        result = [
            seqname,
            source,
            feature,
            int(start),
            int(end),
            score,
            strand,
            frame,
            attributes.replace(';"', '"').replace(
                ";-", "-"
            ),  # correct error in ensemble 78 release
        ]
        entries.append(result)

    entries.sort(key=itemgetter(0, 3, 4))  # seqname, start, end
    new_gff = open(sorted_gff_name, "w")
    for line in header_lines:
        print(line, file=new_gff)
    for entry in entries:
        print("\t".join([str(field) for field in entry]), file=new_gff)
    new_gff.close()
    subprocess.run(["bgzip", "-f", sorted_gff_name], check=True)
    subprocess.run(["tabix", "-p", "gff", compressed_gff_name], check=True)
    logger.warning(f"Created a sorted and tabixed GFF, `{compressed_gff_name}`.")
    return compressed_gff_name


class GFF:
    def __init__(
        self,
        filename,
        dataframe_mode=False,
        store_results=False,
        need_tabix=False,
        feature_type=None,
        gene_exclude_list=None,
        only_protein_coding_genes=False,
    ):
        if not os.path.isfile(filename):
            logger.error(f"Gene model {filename} does not exist!")
            raise SystemExit(1)
        if need_tabix:
            try:
                self.tabix = tabix.open(filename)
                self.tabix.queryi(0, 100, 200)
            except tabix.TabixError:
                logger.warning(f"{filename} not tabixed! Sorting and tabixing new GFF.")
                filename = sort_gff(filename)  # replace provided filename with new file
                self.tabix = tabix.open(filename)
        self.filename = filename
        self.basename = os.path.basename(self.filename)
        self.df = None
        self.entries = None
        self.gene_exclude_list = None
        if gene_exclude_list:
            self.gene_exclude_list = set(
                [item.strip() for item in open(gene_exclude_list, "r").readlines()]
            )
        self.feature_type = feature_type

        if dataframe_mode:
            if self.feature_type:
                self.df = gtfparse.read_gtf(filename, features=[self.feature_type])
            else:
                self.df = gtfparse.read_gtf(filename)
            if self.gene_exclude_list:
                if "gene_name" in self.df.columns:
                    self.df = self.df[
                        self.df["gene_name"] not in self.gene_exclude_list
                    ]
                else:
                    logger.warning(
                        "`gene_name` field missing from GFF; could not filter using provided gene exclude_list."
                    )
            if only_protein_coding_genes:
                if "gene_type" in self.df:  # Gencode
                    self.df = self.df[self.df["gene_type"].str.contains("protein")]
                elif "gene_biotype" in self.df:  # ENSEMBL
                    self.df = self.df[self.df["gene_biotype"].str.contains("protein")]
                else:
                    logger.warning(
                        "Could not isolate protein coding genes. Using all genes."
                    )

        else:
            self.ext = ".".join(self.basename.split(".")[1:])
            self.gzipped = False
            if self.ext.endswith(".gz") or self.ext.endswith(".bgz"):
                self.gzipped = True
                self._handle = gzip.open(filename, "r")
            else:
                self._handle = open(filename, "r")

            self._attr_regexes = [r"(\S+)=(\S+)", r"(\S+) \"(\S+)\""]

            if store_results:
                self.entries = [entry for entry in self]

    def __iter__(self):
        if self.df is not None:
            raise NotImplementedError("iteration in Dataframe mode not implemented")
        return self

    def __next__(self):
        if self.df is not None:
            raise NotImplementedError("iteration in Dataframe mode not implemented")
        while True:
            if self.gzipped:
                line = self._handle.readline().decode("utf-8").strip()
            else:
                line = self._handle.readline().strip()

            if not line:
                raise StopIteration

            if line.startswith("#"):
                continue
            line = line.strip()

            [
                seqname,
                source,
                feature,
                start,
                end,
                score,
                strand,
                frame,
                attributes,
            ] = line.split("\t")

            if self.feature_type and feature != self.feature_type:
                continue

            if self.gene_exclude_list:
                selected_bad_gene = False
                for bad_gene in self.gene_exclude_list:
                    if bad_gene in attributes:
                        selected_bad_gene = True
                        break
                if selected_bad_gene:
                    continue

            result = {
                "seqname": seqname,
                "source": source,
                "feature": feature,
                "start": int(start),
                "end": int(end),
                "score": score,
                "strand": strand,
                "frame": frame,
            }

            attributes = attributes.replace(';"', '"').replace(
                ";-", "-"
            )  # correct error in ensemble 78 release
            for attr_raw in [s.strip() for s in attributes.split(";")]:
                if not attr_raw:
                    continue

                for regex in self._attr_regexes:
                    match = re.match(regex, attr_raw)
                    if match:
                        key, value = match.group(1), match.group(2)
                        result[key.strip()] = value.strip()
            return result

    def sample(self):
        if self.df is None and not self.entries:
            raise NotImplementedError("sample() not implemented in iterator mode")
        if self.df is not None:
            return self.df.sample().to_dict("records")[0]
        return random.choice(self.entries)

    def query(self, chr, start, end):
        if self.df is None and not self.entries:
            raise NotImplementedError("query() not implemented in iterator mode")
        if self.df is not None:
            return self.df.query(
                f"seqname=='{chr}' and start=={start} and end=={end}"
            ).to_dict("records")

        raw_hits = self.tabix.query(chr, start, end)
        hits = []
        for hit in raw_hits:
            if self.gene_exclude_list:
                selected_bad_gene = False
                for bad_gene in self.gene_exclude_list:
                    if bad_gene in hit[8]:
                        selected_bad_gene = True
                        break
                if selected_bad_gene:
                    continue

            result = {
                "seqname": hit[0],
                "source": hit[1],
                "feature": hit[2],
                "start": int(hit[3]),
                "end": int(hit[4]),
                "score": hit[5],
                "strand": hit[6],
                "frame": hit[7],
            }

            attributes = (
                hit[8].replace(';"', '"').replace(";-", "-")
            )  # correct error in ensemble 78 release
            for attr_raw in [s.strip() for s in attributes.split(";")]:
                if not attr_raw or attr_raw == "":
                    continue

                for regex in self._attr_regexes:
                    match = re.match(regex, attr_raw)
                    if match:
                        key, value = match.group(1), match.group(2)
                        result[key.strip()] = value.strip()

            hits.append(result)
        return hits

    def next(self):
        return self.__next__()


class JunctionCache:
    def __init__(self, gff):
        self.gff = gff
        self.exon_starts = defaultdict(SortedList)
        self.exon_ends = defaultdict(SortedList)
        while True:
            try:
                next(self)
            except StopIteration:
                contigs = ", ".join(self.exon_starts.keys())
                logger.debug("Cached " + contigs)
                break

    def __iter__(self):
        return self

    def __next__(self):
        exon = next(self.gff)

        # GFF is 1-based, end inclusive
        # PySam is 0-based, end exclusive
        # starts need to have 1 subtracted
        # ends are already equivelant
        start, end = exon["start"] - 1, exon["end"]
        self.exon_starts[exon["seqname"]].add(start)
        self.exon_ends[exon["seqname"]].add(end)

    def next(self):
        self.__next__()



import csv
import itertools
import logging
import sys

from collections import defaultdict

from ..utils import NGSFile, NGSFileType, GFF

logger = logging.getLogger("strandedness")


def get_filtered_reads_from_region(samfile, gene, min_quality=30, apply_filters=True):
    for read in samfile.fetch(gene["seqname"], gene["start"], gene["end"]):
        if apply_filters and (
            read.is_qcfail
            or read.is_duplicate
            or read.is_secondary
            or read.is_unmapped
            or read.mapq < min_quality
        ):
            continue
        yield read


def disqualify_gene(gene, gff, samfile):
    if gene["seqname"] not in samfile.references:
        return True

    # if there are overlapping features on the positive and negative strand
    # ignore this gene.
    hits = gff.query(gene["seqname"], gene["start"], gene["end"])

    has_positive_gene = None
    has_negative_gene = None

    for hit in hits:
        if hit["feature"] == "gene":
            continue
        if hit["strand"] == "+":
            has_positive_gene = True
        elif hit["strand"] == "-":
            has_negative_gene = True

        if has_positive_gene and has_negative_gene:
            break

    if has_positive_gene and has_negative_gene:
        return True

    return False


def get_reads_rg(read, default="unknown_read_group"):
    for (k, v) in read.tags:
        if k == "RG":
            return v

    return default


def get_predicted_strandedness(forward_evidence_pct, reverse_evidence_pct):
    predicted = "Inconclusive"
    if 0.4 <= forward_evidence_pct and forward_evidence_pct <= 0.6:
        predicted = "Unstranded"
    # This second Unstranded check is redundant with the first check,
    # but more explicit
    elif 0.4 <= reverse_evidence_pct and reverse_evidence_pct <= 0.6:
        predicted = "Unstranded"
    elif 0.8 <= forward_evidence_pct:
        predicted = "Stranded-Forward"
    elif 0.8 <= reverse_evidence_pct:
        predicted = "Stranded-Reverse"

    return predicted


def determine_strandedness(
    ngsfilepath,
    gff,
    n_genes=100,
    min_mapq=30,
    minimum_reads_per_gene=10,
    split_by_rg=False,
    max_iterations_per_try=1000,
    checked_genes=set(),
    overall_evidence=defaultdict(lambda: defaultdict(int)),
):
    try:
        ngsfile = NGSFile(ngsfilepath)
    except FileNotFoundError:
        result = {
            "File": ngsfilepath,
            "TotalReads": "N/A",
            "ForwardPct": "N/A",
            "ReversePct": "N/A",
            "Predicted": "Error opening file.",
        }

        if split_by_rg:
            result["ReadGroup"] = "N/A"

        return [result]

    if ngsfile.filetype != NGSFileType.BAM:
        raise RuntimeError(
            "Invalid file: {}. `strandedness` only supports aligned BAM files!".format(
                ngsfilepath
            )
        )
    samfile = ngsfile.handle

    n_tested_genes = 0
    n_reads_observed = 0
    checked_genes = set()

    read_groups = ["unknown_read_group"]
    if "RG" in samfile.header:
        read_groups += [rg["ID"] for rg in samfile.header["RG"]]

    logger.debug("Starting sampling...")
    total_iterations = 0
    while True:
        total_iterations += 1
        if total_iterations > max_iterations_per_try:
            logger.warn("Max iterations reached! Moving forward with prediction.")
            break

        if n_tested_genes >= n_genes:
            break

        gene = gff.sample()

        if gene["gene_id"] in checked_genes:
            continue

        if disqualify_gene(gene, gff, samfile):
            checked_genes.add(gene["gene_id"])
            continue

        logging.debug("== Candidate Gene ==")
        logging.debug("  [*] ID: {}".format(gene["gene_id"]))
        logging.debug(
            "  [*] Location: {}:{}-{} ({})".format(
                gene["seqname"], gene["start"], gene["end"], gene["strand"]
            )
        )
        logging.debug("  [*] Actions:")

        checked_genes.add(gene["gene_id"])
        relevant_reads = get_filtered_reads_from_region(
            samfile, gene, min_quality=min_mapq
        )

        reads_in_gene = 0
        this_genes_evidence = defaultdict(dict)
        for rg in read_groups:
            this_genes_evidence[rg] = defaultdict(int)

        for read in relevant_reads:
            reads_in_gene += 1
            if read.is_paired:
                if read.is_read1:
                    read_id = "1"
                elif read.is_read2:
                    read_id = "2"
                else:
                    raise RuntimeError("Read is not read 1 or read 2?")
            else:
                # SE reads are equivalent to just assuming the read is read 1.
                read_id = "1"

            if not read.is_reverse:
                read_strand_id = "+"
            else:
                read_strand_id = "-"

            gene_strand_id = gene["strand"]
            reads_observed_state = read_id + read_strand_id + gene_strand_id
            this_reads_rg = get_reads_rg(read)
            this_genes_evidence[this_reads_rg][reads_observed_state] += 1

        if reads_in_gene >= minimum_reads_per_gene:
            logging.debug(
                "    - Sufficient read count ({} >= {})".format(
                    reads_in_gene, minimum_reads_per_gene
                )
            )
            rg_log = ""
            if not this_genes_evidence["unknown_read_group"]:
                this_genes_evidence.pop("unknown_read_group")
            for rg, states in this_genes_evidence.items():
                state_logs = ";".join([f"{state}={n}" for state, n in states.items()])
                rg_log += f"{rg}:{state_logs} "

            logging.debug("    - {}".format(rg_log))

            for rg in this_genes_evidence.keys():
                for state in this_genes_evidence[rg].keys():
                    overall_evidence[rg][state] += this_genes_evidence[rg][state]

            n_tested_genes += 1
            n_reads_observed += reads_in_gene
        else:
            logging.debug(
                "    - Read count too low ({} < {})".format(
                    reads_in_gene, minimum_reads_per_gene
                )
            )

    if split_by_rg:
        results = []
        for rg in read_groups:
            evidence_stranded_forward = (
                overall_evidence[rg]["1++"]
                + overall_evidence[rg]["1--"]
                + overall_evidence[rg]["2+-"]
                + overall_evidence[rg]["2-+"]
            )
            evidence_stranded_reverse = (
                overall_evidence[rg]["1+-"]
                + overall_evidence[rg]["1-+"]
                + overall_evidence[rg]["2++"]
                + overall_evidence[rg]["2--"]
            )
            total_reads = evidence_stranded_forward + evidence_stranded_reverse
            if total_reads == 0 and rg == "unknown_read_group":
                continue

            forward_pct = (
                0
                if total_reads <= 0
                else round(evidence_stranded_forward / total_reads, 4)
            )
            reverse_pct = (
                0
                if total_reads <= 0
                else round(evidence_stranded_reverse / total_reads, 4)
            )
            predicted = get_predicted_strandedness(forward_pct, reverse_pct)

            results.append(
                {
                    "File": ngsfilepath,
                    "ReadGroup": rg,
                    "TotalReads": total_reads,
                    "ForwardPct": forward_pct,
                    "ReversePct": reverse_pct,
                    "Predicted": predicted,
                }
            )
        return (results, checked_genes, overall_evidence)
    else:
        evidence_stranded_forward = 0
        evidence_stranded_reverse = 0

        for rg in read_groups:
            evidence_stranded_forward += (
                overall_evidence[rg]["1++"]
                + overall_evidence[rg]["1--"]
                + overall_evidence[rg]["2+-"]
                + overall_evidence[rg]["2-+"]
            )
            evidence_stranded_reverse += (
                overall_evidence[rg]["1+-"]
                + overall_evidence[rg]["1-+"]
                + overall_evidence[rg]["2++"]
                + overall_evidence[rg]["2--"]
            )
        total_reads = evidence_stranded_forward + evidence_stranded_reverse
        forward_pct = (
            0 if total_reads <= 0 else round(evidence_stranded_forward / total_reads, 4)
        )
        reverse_pct = (
            0 if total_reads <= 0 else round(evidence_stranded_reverse / total_reads, 4)
        )
        predicted = get_predicted_strandedness(forward_pct, reverse_pct)

        return (
            [
                {
                    "File": ngsfilepath,
                    "TotalReads": total_reads,
                    "ForwardPct": forward_pct,
                    "ReversePct": reverse_pct,
                    "Predicted": predicted,
                }
            ],
            checked_genes,
            overall_evidence,
        )


def main(
    ngsfiles,
    gene_model_file,
    outfile=sys.stdout,
    n_genes=100,
    minimum_reads_per_gene=10,
    only_protein_coding_genes=True,
    min_mapq=30,
    split_by_rg=False,
    max_tries=3,
    max_iterations_per_try=1000,
):
    logger.info("Arguments:")
    logger.info("  - Gene model file: {}".format(gene_model_file))
    logger.info("  - Number of genes: {}".format(n_genes))
    logger.info("  - Minimum reads per gene: {}".format(minimum_reads_per_gene))
    logger.info(
        "  - Only consider protein coding genes: {}".format(only_protein_coding_genes)
    )
    logger.info("  - Minimum MAPQ: {}".format(min_mapq))
    logger.info("  - Split by RG: {}".format(split_by_rg))

    if max_iterations_per_try < n_genes:
        logger.error(
            "Max iteration per try cannot be less than number of genes to search!"
        )
        sys.exit(1)

    logger.info("Reading gene model...")
    gff = GFF(
        gene_model_file,
        feature_type="gene",
        store_results=True,
        need_tabix=True,
        only_protein_coding_genes=only_protein_coding_genes,
    )

    logger.info("  - {} features processed.".format(len(gff.entries)))

    fieldnames = ["TotalReads", "ForwardPct", "ReversePct", "Predicted"]
    if split_by_rg:
        fieldnames = ["ReadGroup"] + fieldnames
    fieldnames = ["File"] + fieldnames

    writer = None

    for ngsfilepath in ngsfiles:
        tries_for_file = 0
        checked_genes = set()
        overall_evidence = defaultdict(lambda: defaultdict(int))

        while True:
            tries_for_file += 1
            logger.info("Processing {}, try #{}...".format(ngsfilepath, tries_for_file))

            entries, checked_genes, overall_evidence = determine_strandedness(
                ngsfilepath,
                gff,
                n_genes=n_genes,
                min_mapq=min_mapq,
                minimum_reads_per_gene=minimum_reads_per_gene,
                split_by_rg=split_by_rg,
                max_iterations_per_try=max_iterations_per_try,
                checked_genes=checked_genes,
                overall_evidence=overall_evidence,
            )

            entries_contains_inconclusive = False
            for entry in entries:
                if entry["Predicted"] == "Inconclusive":
                    entries_contains_inconclusive = True
                    break

            if entries_contains_inconclusive and tries_for_file < max_tries:
                continue

            if not writer:
                writer = csv.DictWriter(outfile, fieldnames=fieldnames, delimiter="\t")
                writer.writeheader()
            for entry in entries:
                writer.writerow(entry)
                outfile.flush()

            break



import csv
import itertools
import logging
import pysam
import sys

from collections import defaultdict
from ..utils import NGSFile

logger = logging.getLogger("readlen")


def main(
    ngsfiles,
    outfile=sys.stdout,
    n_samples=100000,
    majority_vote_cutoff=0.7,
):

    writer = csv.DictWriter(
        outfile,
        fieldnames=["File", "Evidence", "MajorityPctDetected", "ConsensusReadLength"],
        delimiter="\t",
    )
    writer.writeheader()
    outfile.flush()

    if n_samples < 1:
        n_samples = None

    for ngsfilepath in ngsfiles:
        read_lengths = defaultdict(int)
        try:
            ngsfile = NGSFile(ngsfilepath)
        except FileNotFoundError:
            result = {
                "File": ngsfilepath,
                "Evidence": "File not found.",
                "MajorityPctDetected": "N/A",
                "ConsensusReadLength": "N/A",
            }

            writer.writerow(result)
            outfile.flush()
            continue

        # accumulate read lengths
        total_reads_sampled = 0
        for read in itertools.islice(ngsfile, n_samples):
            total_reads_sampled += 1
            read_lengths[len(read["query"])] += 1

        read_length_keys_sorted = sorted(
            [int(k) for k in read_lengths.keys()], reverse=True
        )
        putative_max_readlen = read_length_keys_sorted[0]

        # note that simply picking the read length with the highest amount of evidence
        # doesn't make sense things like adapter trimming might shorten the read length,
        # but the read length should never grow past the maximum value.

        # if not, cannot determine, return -1
        pct = read_lengths[putative_max_readlen] / total_reads_sampled
        logger.info("Max read length percentage: {}".format(pct))
        majority_readlen = putative_max_readlen if pct > majority_vote_cutoff else -1

        result = {
            "File": ngsfilepath,
            "Evidence": ";".join(
                ["{}={}".format(k, read_lengths[k]) for k in read_length_keys_sorted]
            ),
            "MajorityPctDetected": round(pct, 4),
            "ConsensusReadLength": majority_readlen,
        }

        writer.writerow(result)
        outfile.flush()



import logging
import sys
import os
import csv

from collections import defaultdict
from pathlib import Path

from ..utils import NGSFile, NGSFileType, GFF, JunctionCache

logger = logging.getLogger("junction-annotation")


def annotate_event(
    found_pos,
    reference_positions,
    fuzzy_range,
):
    for pos in reference_positions.irange(
        found_pos - fuzzy_range, found_pos + fuzzy_range
    ):  # return first match even if multiple found
        return False, pos
    return True, None


def annotate_junctions(
    ngsfilepath,
    cache,
    min_intron,
    min_mapq,
    min_reads,
    fuzzy_range,
    consider_unannotated_references_novel,
    junction_dir,
    disable_junction_files,
):
    try:
        ngsfile = NGSFile(ngsfilepath)
    except FileNotFoundError:
        result = {
            "File": ngsfilepath,
            "total_junctions": "N/A",
            "total_splice_events": "N/A",
            "known_junctions": "N/A",
            "partial_novel_junctions": "N/A",
            "complete_novel_junctions": "N/A",
            "known_spliced_reads": "N/A",
            "complete_novel_spliced_reads": "N/A",
            "partial_novel_spliced_reads": "N/A",
        }
        return [result]

    if ngsfile.filetype != NGSFileType.BAM:
        raise RuntimeError(
            "Invalid file: {}. `junction-annotation` only supports aligned BAM files!".format(
                ngsfilepath
            )
        )
    samfile = ngsfile.handle

    junction_file = None
    if not disable_junction_files:
        junction_filename = os.path.join(junction_dir, ngsfile.basename)
        junction_file = open(f"{junction_filename}.junctions.tsv", "w")
        print(
            "\t".join(
                ["chrom", "intron_start", "intron_end", "read_count", "annotation"]
            ),
            file=junction_file,
        )

    num_known = 0
    num_novel = 0
    num_partial = 0
    num_known_spliced_reads = 0
    num_novel_spliced_reads = 0
    num_partial_spliced_reads = 0

    for contig in samfile.references:
        num_too_few_reads = 0
        logger.info(f"Searching {contig} for splice junctions...")
        found_introns = samfile.find_introns(
            [seg for seg in samfile.fetch(contig) if seg.mapping_quality >= min_mapq]
        )

        events = [
            (intron_start, intron_end, num_reads)
            for (intron_start, intron_end), num_reads in found_introns.items()
            if intron_end - intron_start >= min_intron
        ]
        if not events:
            logger.debug(
                f"No valid splice junctions in {contig}. {len(found_introns)} potential junctions too short."
            )
            continue
        logger.debug(
            f"Found {len(events)} potential splice junctions. {len(found_introns) - len(events)} potential junctions too short."
        )

        if contig not in cache.exon_starts:
            logger.info(f"{contig} not found in GFF. All events marked `unannotated_reference`.")
            annotation = "unannotated_reference"
            if consider_unannotated_references_novel:
                logger.info(f"Events being considered novel for summary report.")

            for intron_start, intron_end, num_reads in events:
                if num_reads < min_reads:
                    num_too_few_reads += 1
                    continue
                if consider_unannotated_references_novel:
                    num_novel += 1
                    num_novel_spliced_reads += num_reads

                if junction_file:
                    print(
                        "\t".join(
                            [
                                contig,
                                str(intron_start),
                                str(intron_end),
                                str(num_reads),
                                annotation,
                            ]
                        ),
                        file=junction_file,
                    )

            logger.debug(
                f"{num_too_few_reads} potential junctions didn't have enough read support."
            )
            logger.debug(f"{len(events) - num_too_few_reads} junctions annotated.")
            continue

        collapsed_junctions = defaultdict(int)

        for intron_start, intron_end, num_reads in events:
            start_novel, ref_start = annotate_event(
                intron_start, cache.exon_ends[contig], fuzzy_range
            )

            end_novel, ref_end = annotate_event(
                intron_end, cache.exon_starts[contig], fuzzy_range
            )

            if ref_start:
                start = ref_start
            else:
                start = intron_start
            if ref_end:
                end = ref_end
            else:
                end = intron_end

            # if fuzzy searching, collapse these reads into nearby events
            if fuzzy_range:
                collapsed_junctions[(start, end)] += num_reads
            # if not fuzzy searching, tally reads and write to outfile
            else:
                if num_reads < min_reads:
                    num_too_few_reads += 1
                    continue
                if start_novel and end_novel:
                    annotation = "complete_novel"
                    num_novel += 1
                    num_novel_spliced_reads += num_reads
                elif start_novel or end_novel:
                    annotation = "partial_novel"
                    num_partial += 1
                    num_partial_spliced_reads += num_reads
                else:
                    annotation = "annotated"
                    num_known += 1
                    num_known_spliced_reads += num_reads

                if junction_file:
                    print(
                        "\t".join(
                            [
                                contig,
                                str(start),
                                str(end),
                                str(num_reads),
                                annotation,
                            ]
                        ),
                        file=junction_file,
                    )

        # if not fuzzy searching, collapsed_junctions is empty and loop is skipped,
        # reads will have been tallied and written to outfile already
        for (intron_start, intron_end), num_reads in sorted(
            collapsed_junctions.items()
        ):
            if num_reads < min_reads:
                num_too_few_reads += 1
                continue
            start_novel, _ = annotate_event(intron_start, cache.exon_ends[contig], 0)

            end_novel, _ = annotate_event(intron_end, cache.exon_starts[contig], 0)

            if start_novel and end_novel:
                annotation = "complete_novel"
                num_novel += 1
                num_novel_spliced_reads += num_reads
            elif start_novel or end_novel:
                annotation = "partial_novel"
                num_partial += 1
                num_partial_spliced_reads += num_reads
            else:
                annotation = "annotated"
                num_known += 1
                num_known_spliced_reads += num_reads

            if not disable_junction_files:
                print(
                    "\t".join(
                        [
                            contig,
                            str(intron_start),
                            str(intron_end),
                            str(num_reads),
                            annotation,
                        ]
                    ),
                    file=junction_file,
                )

        logger.debug(
            f"{num_too_few_reads} potential junctions didn't have enough read support."
        )
        if not fuzzy_range:
            logger.debug(f"{len(events) - num_too_few_reads} junctions annotated.")
        else:
            logger.debug(
                f"{len(collapsed_junctions) - num_too_few_reads} junctions annotated."
            )
    if not disable_junction_files:
        junction_file.close()

    result = {
        "File": ngsfilepath,
        "total_junctions": num_known + num_novel + num_partial,
        "total_splice_events": num_known_spliced_reads
        + num_novel_spliced_reads
        + num_partial_spliced_reads,
        "known_junctions": num_known,
        "partial_novel_junctions": num_partial,
        "complete_novel_junctions": num_novel,
        "known_spliced_reads": num_known_spliced_reads,
        "partial_novel_spliced_reads": num_partial_spliced_reads,
        "complete_novel_spliced_reads": num_novel_spliced_reads,
    }
    return result


def main(
    ngsfiles,
    gene_model_file,
    outfile=sys.stdout,
    min_intron=50,
    min_mapq=30,
    min_reads=2,
    fuzzy_range=0,
    consider_unannotated_references_novel=False,
    junction_dir="./",
    disable_junction_files=False,
):
    logger.info("Arguments:")
    logger.info("  - Gene model file: {}".format(gene_model_file))
    logger.info("  - Minimum intron length: {}".format(min_intron))
    logger.info("  - Minimum MAPQ: {}".format(min_mapq))
    logger.info("  - Minimum reads per junction: {}".format(min_reads))
    logger.info("  - Fuzzy junction range: +-{}".format(fuzzy_range))
    logger.info("  - Consider unannotated references novel: {}".format(consider_unannotated_references_novel))
    if not disable_junction_files:
        logger.info("  - Junction file directory: {}".format(junction_dir))
    else:
        logger.info("  - Junction file directory: <disabled>")

    logger.info("Processing gene model...")
    gff = GFF(
        gene_model_file,
        feature_type="exon",
        dataframe_mode=False,
    )
    cache = JunctionCache(gff)
    logger.info("Done")

    junction_dir = Path(junction_dir)
    if not disable_junction_files:
        junction_dir.mkdir(parents=True, exist_ok=True)

    writer = None
    for ngsfilepath in ngsfiles:
        entry = annotate_junctions(
            ngsfilepath,
            cache,
            min_intron=min_intron,
            min_mapq=min_mapq,
            min_reads=min_reads,
            fuzzy_range=fuzzy_range,
            consider_unannotated_references_novel=consider_unannotated_references_novel,
            junction_dir=junction_dir,
            disable_junction_files=disable_junction_files,
        )

        if not writer:
            fieldnames = [
                "File",
                "total_junctions",
                "total_splice_events",
                "known_junctions",
                "partial_novel_junctions",
                "complete_novel_junctions",
                "known_spliced_reads",
                "partial_novel_spliced_reads",
                "complete_novel_spliced_reads",
            ]

            writer = csv.DictWriter(outfile, fieldnames=fieldnames, delimiter="\t")
            writer.writeheader()
        writer.writerow(entry)
        outfile.flush()



import csv
import itertools
import pysam
import sys

import logging
from collections import defaultdict

from ..utils import NGSFile, NGSFileType

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# sets are created by doing PHRED+33 decoding of each encoding's ASCII ranges
SANGER_SET = set([i for i in range(0, 93)])
ILLUMINA_1_0_SET = set([i for i in range(26, 93)])
ILLUMINA_1_3_SET = set([i for i in range(31, 93)])


def main(ngsfiles, outfile=sys.stdout, n_samples=1000000):

    writer = csv.DictWriter(
        outfile,
        fieldnames=["File", "Evidence", "ProbableEncoding"],
        delimiter="\t",
    )
    writer.writeheader()
    outfile.flush()

    if n_samples < 1:
        n_samples = None

    for ngsfilepath in ngsfiles:
        try:
            ngsfile = NGSFile(ngsfilepath, store_qualities=True)
        except FileNotFoundError:
            result = {
                "File": ngsfilepath,
                "Evidence": "File not found.",
                "ProbableEncoding": "N/A",
            }
            writer.writerow(result)
            outfile.flush()
            continue

        score_set = set()
        for read in itertools.islice(ngsfile, n_samples):
            score_set.update(read["quality"])

        highest_ascii = str(max(score_set) + 33)
        lowest_ascii = str(min(score_set) + 33)
        if score_set <= ILLUMINA_1_3_SET:
            result = {
                "File": ngsfilepath,
                "Evidence": "ASCII range: {}-{}".format(lowest_ascii, highest_ascii),
                "ProbableEncoding": "Illumina 1.3",
            }
        elif score_set <= ILLUMINA_1_0_SET:
            result = {
                "File": ngsfilepath,
                "Evidence": "ASCII range: {}-{}".format(lowest_ascii, highest_ascii),
                "ProbableEncoding": "Solexa/Illumina 1.0",
            }
        elif score_set <= SANGER_SET:
            result = {
                "File": ngsfilepath,
                "Evidence": "ASCII range: {}-{}".format(lowest_ascii, highest_ascii),
                "ProbableEncoding": "Sanger/Illumina 1.8",
            }
        else:
            result = {
                "File": ngsfilepath,
                "Evidence": "ASCII values outside known PHRED encoding ranges: {}-{}".format(
                    lowest_ascii, highest_ascii
                ),
                "ProbableEncoding": "Unknown",
            }

        writer.writerow(result)
        outfile.flush()



import csv
import itertools
import pysam
import re
import sys

import logging
from collections import defaultdict

from ..utils import NGSFile

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

instrument_ids = {
    "^HWI-M[0-9]{4}$": ["MiSeq"],
    "^HWUSI": ["Genome Analyzer IIx"],
    "^M[0-9]{5}$": ["MiSeq"],
    "^HWI-C[0-9]{5}$": ["HiSeq 1500"],
    "^C[0-9]{5}$": ["HiSeq 1500"],
    "^HWI-ST[0-9]{3,5}(_[0-9]{9})?$": ["HiSeq 2000"],
    "^HWI-D[0-9]{5}$": ["HiSeq 2000", "HiSeq 2500"],
    "^A[0-9]{5}$": ["NovaSeq"],
    "^D[0-9]{5}$": ["HiSeq 2500"],
    "^J[0-9]{5}$": ["HiSeq 3000"],
    "^K[0-9]{5}$": ["HiSeq 3000", "HiSeq 4000"],
    "^E[0-9]{5}$": ["HiSeq X"],
    "^N[0-9]{5}$": ["NextSeq"],
    "^NB[0-9]{6}$": ["NextSeq"],
    "^NS[0-9]{6}$": ["NextSeq"],
    "^MN[0-9]{5}$": ["MiniSeq"],
}

flowcell_ids = {
    "^C[A-Z0-9]{4}ANXX$": [
        "HiSeq 1500",
        "HiSeq 2000",
        "HiSeq 2500",
    ],  # High Output (8-lane) v4 flow cell
    "^C[A-Z0-9]{4}ACXX$": [
        "HiSeq 1000",
        "HiSeq 1500",
        "HiSeq 2000",
        "HiSeq 2500",
    ],  # High Output (8-lane) v3 flow cell
    "^D[A-Z0-9]{4}ACXX$": [
        "HiSeq 1000",
        "HiSeq 1500",
        "HiSeq 2000",
        "HiSeq 2500",
    ],  # High Output (8-lane) v3 flow cell
    "^H[A-Z0-9]{4}ADXX$": [
        "HiSeq 1500",
        "HiSeq 2000",
        "HiSeq 2500",
    ],  # Rapid Run (2-lane) v1 flow cell
    "^H[A-Z0-9]{4}BCXX$": [
        "HiSeq 1500",
        "HiSeq 2500",
    ],  # Rapid Run (2-lane) v2 flow cell
    "^H[A-Z0-9]{4}BCXY$": [
        "HiSeq 1500",
        "HiSeq 2500",
    ],  # Rapid Run (2-lane) v2 flow cell
    "^H[A-Z0-9]{4}BBXX$": ["HiSeq 4000"],  # (8-lane) v1 flow cell
    "^H[A-Z0-9]{4}BBXY$": ["HiSeq 4000"],  # (8-lane) v1 flow cell
    "^H[A-Z0-9]{4}CCXX$": ["HiSeq X"],  # (8-lane) flow cell
    "^H[A-Z0-9]{4}CCXY$": ["HiSeq X"],  # (8-lane) flow cell
    "^H[A-Z0-9]{4}ALXX$": ["HiSeq X"],  # (8-lane) flow cell
    "^H[A-Z0-9]{4}BGX[A-Z,0-9]$": ["NextSeq"],  # High output flow cell
    "^H[A-Z0-9]{4}AFXX$": ["NextSeq"],  # Mid output flow cell
    "^H[A-Z0-9]{5}RXX$": ["NovaSeq"],  # S1 flow cell
    "^H[A-Z0-9]{5}RXX$": ["NovaSeq"],  # SP flow cell
    "^H[A-Z0-9]{5}MXX$": ["NovaSeq"],  # S2 flow cell
    "^H[A-Z0-9]{5}SXX$": ["NovaSeq"],  # S4 flow cell
    "^A[A-Z0-9]{4}$": ["MiSeq"],  # MiSeq flow cell
    "^B[A-Z0-9]{4}$": ["MiSeq"],  # MiSeq flow cell
    "^D[A-Z0-9]{4}$": ["MiSeq"],  # MiSeq nano flow cell
    # "^D[A-Z0-9]{4}$" : ["HiSeq 2000", "HiSeq 2500"],  # Unknown HiSeq flow cell examined in SJ data
    "^G[A-Z0-9]{4}$": ["MiSeq"],  # MiSeq micro flow cell
}

upgrade_sets = [(set(["HiSeq 2000", "HiSeq 2500"]), ["HiSeq 2000", "HiSeq 2500"])]


def derive_instrument_from_iid(iid):
    matching_instruments = set()

    for pattern in instrument_ids.keys():
        if re.search(pattern, iid):
            matching_instruments |= set(instrument_ids[pattern])

    return matching_instruments


def predict_instrument_from_iids(iids):
    possible_instruments_by_iid = set()
    detected_at_least_one_instrument = False

    for iid in iids:
        if len(possible_instruments_by_iid) == 0:
            possible_instruments_by_iid = derive_instrument_from_iid(iid)
            detected_at_least_one_instrument = True
        else:
            possible_instruments_by_iid &= derive_instrument_from_iid(iid)

    return possible_instruments_by_iid, detected_at_least_one_instrument


def derive_instrument_from_fcid(fcid):
    matching_instruments = set()

    for pattern in flowcell_ids.keys():
        if re.search(pattern, fcid):
            matching_instruments |= set(flowcell_ids[pattern])

    return matching_instruments


def predict_instrument_from_fcids(fcids):
    possible_instruments_by_fcid = set()
    detected_at_least_one_instrument = False

    for fcid in fcids:
        if len(possible_instruments_by_fcid) == 0:
            possible_instruments_by_fcid = derive_instrument_from_fcid(fcid)
            detected_at_least_one_instrument = True
        else:
            possible_instruments_by_fcid &= derive_instrument_from_fcid(fcid)

    return possible_instruments_by_fcid, detected_at_least_one_instrument


def resolve_instrument(
    possible_instruments_by_iid,
    possible_instruments_by_fcid,
    at_least_one_instrument_detected,
    malformed_read_names_detected,
):
    if len(possible_instruments_by_iid) == 0 and len(possible_instruments_by_fcid) == 0:
        if at_least_one_instrument_detected:
            return (
                set(["multiple instruments"]),
                "unknown confidence",
                "multiple instruments were detected in this sample",
            )
        elif malformed_read_names_detected:
            return (
                set(["unknown"]),
                "no confidence",
                "read names not in Illumina format",
            )
        else:
            return set(["unknown"]), "no confidence", "no match"

    if len(possible_instruments_by_iid) == 0:
        if not malformed_read_names_detected:
            confidence = "medium confidence"
            if len(possible_instruments_by_fcid) > 1:
                confidence = "low confidence"
            return possible_instruments_by_fcid, confidence, "flowcell id"
        else:
            confidence = "low confidence"
            return possible_instruments_by_fcid, confidence, "flowcell id"

    if len(possible_instruments_by_fcid) == 0:
        if not malformed_read_names_detected:
            confidence = "medium confidence"
            if len(possible_instruments_by_iid) > 1:
                confidence = "low confidence"
            return possible_instruments_by_iid, confidence, "instrument id"
        else:
            confidence = "low confidence"
            return (
                possible_instruments_by_iid,
                confidence,
                "instrument id",
            )

    overlapping_instruments = possible_instruments_by_iid & possible_instruments_by_fcid
    if len(overlapping_instruments) == 0:
        return (
            set(["conflicting evidence"]),
            "high confidence",
            "Case needs triaging: {} by iid, {} by fcid".format(
                " or ".join(possible_instruments_by_iid),
                " or ".join(possible_instruments_by_fcid),
            ),
        )
    else:
        return (
            overlapping_instruments,
            "high confidence",
            "instrument id and flowcell id",
        )


def main(ngsfiles, outfile=sys.stdout, n_samples=10000):
    writer = csv.DictWriter(
        outfile,
        fieldnames=["File", "Instrument", "Confidence", "Basis"],
        delimiter="\t",
    )
    writer.writeheader()
    outfile.flush()

    if n_samples < 1:
        n_samples = None

    for ngsfilepath in ngsfiles:
        try:
            ngsfile = NGSFile(ngsfilepath)
        except FileNotFoundError:
            result = {
                "File": ngsfilepath,
                "Instrument": "File not found.",
                "Confidence": "N/A",
                "Basis": "N/A",
            }

            writer.writerow(result)
            outfile.flush()
            continue

        instruments = set()
        flowcells = set()
        malformed_read_names = False

        # accumulate instrument and flowcell IDs
        for read in itertools.islice(ngsfile, n_samples):
            parts = read["query_name"].split(":")
            if len(parts) != 7:  # not Illumina format
                malformed_read_names = True
                iid = parts[0]  # attempt to recover machine name
                instruments.add(iid)
                for rg in ngsfile.handle.header.to_dict()["RG"]:
                    if rg["ID"] == read["read_group"]:
                        if "PU" in rg:
                            flowcells.add(rg["PU"])
                        if "PM" in rg:
                            instruments.add(rg["PM"])
                continue
            iid, fcid = parts[0], parts[2]
            instruments.add(iid)
            flowcells.add(fcid)
        if malformed_read_names:
            logger.warning(
                "Encountered read names not in Illumina format. Recovery attempted."
            )
        (
            possible_instruments_by_iid,
            detected_instrument_by_iid,
        ) = predict_instrument_from_iids(instruments)
        (
            possible_instruments_by_fcid,
            detected_instrument_by_fcid,
        ) = predict_instrument_from_fcids(flowcells)

        instruments, confidence, based_on = resolve_instrument(
            possible_instruments_by_iid,
            possible_instruments_by_fcid,
            detected_instrument_by_iid | detected_instrument_by_fcid,
            malformed_read_names,
        )
        for upgrade_set in upgrade_sets:
            if instruments.issubset(upgrade_set[0]):
                instruments = upgrade_set[1]
                break

        result = {
            "File": ngsfilepath,
            "Instrument": " or ".join(instruments),
            "Confidence": confidence,
            "Basis": based_on,
        }

        writer.writerow(result)
        outfile.flush()



#!/usr/bin/env python3

import argparse
import csv
import logging
import sys

from ngsderive import utils
from ngsderive.commands import (
    readlen,
    instrument,
    strandedness,
    encoding,
    junction_annotation,
)

logger = logging.getLogger("ngsderive")


def get_args():
    class SaneFormatter(
        argparse.RawTextHelpFormatter, argparse.ArgumentDefaultsHelpFormatter
    ):
        pass

    parser = argparse.ArgumentParser(
        description="Derive useful information (or best guess) from next-generation sequencing files.",
        formatter_class=SaneFormatter,
    )
    subparsers = parser.add_subparsers(dest="subcommand")

    common = argparse.ArgumentParser(add_help=False, formatter_class=SaneFormatter)
    common.add_argument(
        "ngsfiles",
        type=str,
        nargs="+",
        help="Next-generation sequencing files to process (BAM or FASTQ).",
    )
    common.add_argument(
        "-o",
        "--outfile",
        type=str,
        help="Write to filename rather than standard out.",
        default="stdout",
    )
    common.add_argument(
        "--debug", default=False, action="store_true", help="Enable DEBUG log level."
    )
    common.add_argument(
        "-v",
        "--verbose",
        default=False,
        action="store_true",
        help="Enable INFO log level.",
    )
    common.add_argument("--version", action="version", version="%(prog)s 1.0.0")

    readlen = subparsers.add_parser(
        "readlen", parents=[common], formatter_class=SaneFormatter
    )
    readlen.add_argument(
        "-c",
        "--majority-vote-cutoff",
        type=float,
        help="To call a majority readlen, the maximum read length must have at least `majority-vote-cutoff`%% reads in support.",
        default=0.7,
    )
    readlen.add_argument(
        "-n",
        "--n-samples",
        type=int,
        help="How many reads to sample. Any n < 1 to parse whole file.",
        default=1000000,
    )

    instrument = subparsers.add_parser(
        "instrument", parents=[common], formatter_class=SaneFormatter
    )
    instrument.add_argument(
        "-n",
        "--n-samples",
        type=int,
        help="How many reads to sample. Any n < 1 to parse whole file.",
        default=10000,
    )

    strandedness = subparsers.add_parser(
        "strandedness", parents=[common], formatter_class=SaneFormatter
    )
    strandedness.add_argument(
        "-g", "--gene-model", help="Gene model as a GFF/GTF file.", required=True
    )
    strandedness.add_argument(
        "--max-tries",
        type=int,
        default=3,
        help="When inconclusive, the test will repeat until this many tries have been reached. "
             "Evidence of previous attempts is saved and reused, leading to a larger sample size with multiple attempts.",
    )
    strandedness.add_argument(
        "--max-iterations-per-try",
        type=int,
        default=None,
        help="At most, search this many times for genes that satisfy our search criteria. Default is 10 * n-genes.",
    )
    strandedness.add_argument(
        "-m",
        "--minimum-reads-per-gene",
        type=int,
        help="Filter any genes that don't have at least `m` reads.",
        default=10,
    )
    strandedness.add_argument(
        "-n", "--n-genes", type=int, help="How many genes to sample.", default=1000
    )
    strandedness.add_argument(
        "-q",
        "--min-mapq",
        type=int,
        help="Minimum MAPQ to consider for reads.",
        default=30,
    )
    protein_coding_parser = strandedness.add_mutually_exclusive_group(required=False)
    protein_coding_parser.add_argument(
        "--only-protein-coding-genes",
        dest="only_protein_coding_genes",
        action="store_true",
        help="Only consider protein coding genes",
    )
    protein_coding_parser.add_argument(
        "--no-only-protein-coding-genes",
        dest="only_protein_coding_genes",
        action="store_false",
    )
    split_by_rg_parser = strandedness.add_mutually_exclusive_group(required=False)
    split_by_rg_parser.add_argument(
        "--split-by-rg",
        dest="split_by_rg",
        action="store_true",
        help="Contain one entry per read group.",
    )
    split_by_rg_parser.add_argument(
        "--no-split-by-rg", dest="split_by_rg", action="store_false"
    )
    strandedness.set_defaults(only_protein_coding_genes=True, split_by_rg=False)

    encoding = subparsers.add_parser(
        "encoding", parents=[common], formatter_class=SaneFormatter
    )
    encoding.add_argument(
        "-n",
        "--n-samples",
        type=int,
        help="How many reads to sample. Any n < 1 to parse whole file.",
        default=1000000,
    )

    junction_annotation = subparsers.add_parser(
        "junction-annotation", parents=[common], formatter_class=SaneFormatter
    )
    junction_annotation.add_argument(
        "-g", "--gene-model", help="Gene model as a GFF/GTF file.", required=True
    )
    junction_annotation.add_argument(
        "-j",
        "--junction-files-dir",
        help="Directory to write annotated junction files to.",
        default="./",
    )
    junction_annotation.add_argument(
        "-d",
        "--disable-junction-files",
        help="Disable generating junction files.",
        action="store_true",
    )
    junction_annotation.add_argument(
        "-i",
        "--min-intron",
        type=int,
        help="Minimum size of intron to be considered a splice.",
        default=50,
    )
    junction_annotation.add_argument(
        "-q",
        "--min-mapq",
        type=int,
        help="Minimum MAPQ to consider for supporting reads.",
        default=30,
    )
    junction_annotation.add_argument(
        "-m",
        "--min-reads",
        type=int,
        help="Filter any junctions that don't have at least `m` reads.",
        default=2,
    )
    junction_annotation.add_argument(
        "-k",
        "--fuzzy-junction-match-range",
        type=int,
        help="Consider found splices within `+-k` bases of a known splice event annotated.",
        default=0,
    )
    junction_annotation.add_argument(
        "-c",
        "--consider-unannotated-references-novel",
        action="store_true",
        default=False,
        help="For the summary report, consider all events on unannotated reference sequences `complete_novel`. "
             "Default is to exclude them from the summary. "
             "Either way, they will be annotated as `unannotated_reference` in the junctions file.",
    )

    args = parser.parse_args()
    if not args.subcommand:
        parser.print_help()
        sys.exit(1)

    return args


def setup_logging(log_level=logging.INFO):
    """Set up the logging.

    Forked from MIT code here: https://github.com/MisterWil/abodepy.
    """
    logging.basicConfig(level=log_level)
    fmt = "%(asctime)s %(levelname)s (%(threadName)s) " "[%(name)s] %(message)s"
    colorfmt = "%(log_color)s{}%(reset)s".format(fmt)
    datefmt = "%H:%M:%S"

    try:
        from colorlog import ColoredFormatter

        logging.getLogger().handlers[0].setFormatter(
            ColoredFormatter(
                colorfmt,
                datefmt=datefmt,
                reset=True,
                log_colors={
                    "DEBUG": "cyan",
                    "INFO": "green",
                    "WARNING": "yellow",
                    "ERROR": "red",
                    "CRITICAL": "red",
                },
            )
        )
    except ImportError:
        pass

    logger = logging.getLogger("")
    logger.setLevel(log_level)


def process_args(args):
    # setup logging
    log_level = logging.WARN

    if args.verbose:
        log_level = logging.INFO

    if args.debug:
        log_level = logging.DEBUG

    setup_logging(log_level)

    # set output file
    if args.outfile == "stdout":
        args.outfile = sys.stdout
    else:
        args.outfile = open(args.outfile, "w")


def run():
    args = get_args()
    process_args(args)

    if args.subcommand == "readlen":
        readlen.main(
            args.ngsfiles,
            outfile=args.outfile,
            n_samples=args.n_samples,
            majority_vote_cutoff=args.majority_vote_cutoff,
        )
    if args.subcommand == "instrument":
        instrument.main(
            args.ngsfiles,
            outfile=args.outfile,
            n_samples=args.n_samples,
        )
    if args.subcommand == "strandedness":
        max_iters = args.max_iterations_per_try
        if not max_iters:
            max_iters = 10 * args.n_genes
        strandedness.main(
            args.ngsfiles,
            args.gene_model,
            outfile=args.outfile,
            n_genes=args.n_genes,
            minimum_reads_per_gene=args.minimum_reads_per_gene,
            only_protein_coding_genes=args.only_protein_coding_genes,
            min_mapq=args.min_mapq,
            split_by_rg=args.split_by_rg,
            max_iterations_per_try=max_iters,
        )
    if args.subcommand == "encoding":
        encoding.main(
            args.ngsfiles,
            outfile=args.outfile,
            n_samples=args.n_samples,
        )
    if args.subcommand == "junction-annotation":
        junction_annotation.main(
            args.ngsfiles,
            args.gene_model,
            outfile=args.outfile,
            min_intron=args.min_intron,
            min_mapq=args.min_mapq,
            min_reads=args.min_reads,
            fuzzy_range=args.fuzzy_junction_match_range,
            consider_unannotated_references_novel=args.consider_unannotated_references_novel,
            junction_dir=args.junction_files_dir,
            disable_junction_files=args.disable_junction_files,
        )
